import type { FontData, MathValueRecord } from "@math-blocks/opentype";
import type { Context } from "./types";
export declare type Dist = number;
export declare type Dim = Readonly<{
    readonly width: Dist;
    readonly depth: Dist;
    readonly height: Dist;
}>;
declare type Style = {
    readonly color?: string;
    readonly cancel?: number;
};
declare type Common = {
    readonly id?: number;
    readonly style: Style;
};
export declare type Content = {
    readonly type: "static";
    readonly nodes: readonly Node[];
} | {
    readonly type: "cursor";
    readonly left: readonly Node[];
    readonly right: readonly Node[];
} | {
    readonly type: "selection";
    readonly left: readonly Node[];
    readonly selection: readonly Node[];
    readonly right: readonly Node[];
};
export declare type HBox = {
    readonly type: "HBox";
    readonly shift: Dist;
    readonly content: Content;
    readonly fontSize: number;
} & Common & Dim;
export declare type VBox = {
    readonly type: "VBox";
    readonly shift: Dist;
    readonly content: readonly Node[];
    readonly fontSize: number;
} & Common & Dim;
export declare type Glyph = {
    readonly type: "Glyph";
    readonly char?: string;
    readonly glyphID: number;
    readonly size: number;
    readonly fontData: FontData;
    readonly pending?: boolean;
} & Common;
export declare type Kern = {
    readonly type: "Kern";
    readonly size: Dist;
    readonly flag?: "start" | "end";
} & Common;
export declare type HRule = {
    readonly type: "HRule";
    readonly thickness: number;
    readonly width: number;
} & Common;
export declare type Node = HBox | VBox | Glyph | Kern | HRule;
export declare const rebox: (box: HBox, before: Kern, after: Kern) => HBox;
export declare const makeKern: (size: Dist, flag?: "start" | "end" | undefined) => Kern;
export declare const makeHRule: (thickness: number, width: number) => HRule;
export declare const makeGlyph: (char: string, glyphID: number, context: Context) => Glyph;
export declare const getCharAdvance: (glyph: Glyph) => number;
export declare const getCharBearingX: (glyph: Glyph) => number;
export declare const getCharWidth: (glyph: Glyph) => number;
export declare const getCharHeight: (glyph: Glyph) => number;
export declare const getCharDepth: (glyph: Glyph) => number;
export declare const getWidth: (node: Node) => number;
export declare const getHeight: (node: Node) => number;
export declare const getDepth: (node: Node) => number;
export declare const vsize: (node: Node) => number;
export declare const hlistWidth: (nodes: readonly Node[]) => number;
export declare const makeStaticHBox: (nodes: readonly Node[], context: Context) => HBox;
export declare const makeCursorHBox: (left: readonly Node[], right: readonly Node[], context: Context) => HBox;
export declare const makeSelectionHBox: (left: readonly Node[], selection: readonly Node[], right: readonly Node[], context: Context) => HBox;
export declare const makeVBox: (width: Dist, node: Node, upList: readonly Node[], dnList: readonly Node[], context: Context) => VBox;
export declare const getConstantValue: (constant: MathValueRecord, context: Context) => number;
export {};
