"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getConstantValue = exports.makeVBox = exports.makeSelectionHBox = exports.makeCursorHBox = exports.makeStaticHBox = exports.hlistWidth = exports.vsize = exports.getDepth = exports.getHeight = exports.getWidth = exports.getCharDepth = exports.getCharHeight = exports.getCharWidth = exports.getCharBearingX = exports.getCharAdvance = exports.makeGlyph = exports.makeHRule = exports.makeKern = exports.rebox = void 0;
var core_1 = require("@math-blocks/core");
var utils_1 = require("./utils");
var makeHBox = function (dim, content, context) {
    return __assign(__assign({ type: "HBox" }, dim), { shift: 0, content: content, fontSize: (0, utils_1.fontSizeForContext)(context), style: {} });
};
var rebox = function (box, before, after) {
    if (box.content.type === "static") {
        return __assign(__assign({}, box), { width: before.size + box.width + after.size, content: __assign(__assign({}, box.content), { nodes: __spreadArray(__spreadArray([before], __read(box.content.nodes), false), [after], false) }) });
    }
    else {
        return __assign(__assign({}, box), { width: before.size + box.width + after.size, content: __assign(__assign({}, box.content), { left: __spreadArray([before], __read(box.content.left), false), right: __spreadArray(__spreadArray([], __read(box.content.right), false), [after], false) }) });
    }
};
exports.rebox = rebox;
var makeKern = function (size, flag) { return ({
    type: "Kern",
    size: size,
    flag: flag,
    style: {},
}); };
exports.makeKern = makeKern;
var makeHRule = function (thickness, width) { return ({
    type: "HRule",
    thickness: thickness,
    width: width,
    style: {},
}); };
exports.makeHRule = makeHRule;
var makeGlyph = function (char, glyphID, context) {
    return {
        type: "Glyph",
        char: char,
        glyphID: glyphID,
        size: (0, utils_1.fontSizeForContext)(context),
        fontData: context.fontData,
        style: {},
    };
};
exports.makeGlyph = makeGlyph;
var getCharAdvance = function (glyph) {
    var font = glyph.fontData.font;
    var metrics = font.getGlyphMetrics(glyph.glyphID);
    var unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error("metrics do not exist for \"" + glyph.char + "\"");
    }
    return (metrics.advance * glyph.size) / unitsPerEm;
};
exports.getCharAdvance = getCharAdvance;
var getCharBearingX = function (glyph) {
    var font = glyph.fontData.font;
    var metrics = font.getGlyphMetrics(glyph.glyphID);
    var unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error("metrics do not exist for \"" + glyph.char + "\"");
    }
    return (metrics.bearingX * glyph.size) / unitsPerEm;
};
exports.getCharBearingX = getCharBearingX;
var getCharWidth = function (glyph) {
    var font = glyph.fontData.font;
    var metrics = font.getGlyphMetrics(glyph.glyphID);
    var unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error("metrics do not exist for \"" + glyph.char + "\"");
    }
    return (metrics.width * glyph.size) / unitsPerEm;
};
exports.getCharWidth = getCharWidth;
var getCharHeight = function (glyph) {
    var font = glyph.fontData.font;
    var metrics = font.getGlyphMetrics(glyph.glyphID);
    var unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error("metrics do not exist for \"" + glyph.char + "\"");
    }
    return (metrics.bearingY * glyph.size) / unitsPerEm;
};
exports.getCharHeight = getCharHeight;
var getCharDepth = function (glyph) {
    var font = glyph.fontData.font;
    var metrics = font.getGlyphMetrics(glyph.glyphID);
    var unitsPerEm = font.head.unitsPerEm;
    if (!metrics) {
        throw new Error("metrics do not exist for \"" + glyph.char + "\"");
    }
    return ((metrics.height - metrics.bearingY) * glyph.size) / unitsPerEm;
};
exports.getCharDepth = getCharDepth;
var getWidth = function (node) {
    switch (node.type) {
        case "HBox":
            return node.width;
        case "VBox":
            return node.width;
        case "Glyph":
            return (0, exports.getCharAdvance)(node);
        case "Kern":
            return node.size;
        case "HRule":
            return node.width;
        default:
            throw new core_1.UnreachableCaseError(node);
    }
};
exports.getWidth = getWidth;
var getHeight = function (node) {
    switch (node.type) {
        case "HBox":
            return node.height - node.shift;
        case "VBox":
            return node.height - node.shift;
        case "Glyph":
            return (0, exports.getCharHeight)(node);
        case "Kern":
            return 0;
        case "HRule":
            return node.thickness / 2;
        default:
            throw new core_1.UnreachableCaseError(node);
    }
};
exports.getHeight = getHeight;
var getDepth = function (node) {
    switch (node.type) {
        case "HBox":
            return node.depth + node.shift;
        case "VBox":
            return node.depth + node.shift;
        case "Glyph":
            return (0, exports.getCharDepth)(node);
        case "Kern":
            return 0;
        case "HRule":
            return node.thickness / 2;
        default:
            throw new core_1.UnreachableCaseError(node);
    }
};
exports.getDepth = getDepth;
var vsize = function (node) {
    switch (node.type) {
        case "HBox":
            return node.height + node.depth;
        case "VBox":
            return node.height + node.depth;
        case "Glyph":
            return (0, exports.getCharHeight)(node) + (0, exports.getCharDepth)(node);
        case "Kern":
            return node.size;
        case "HRule":
            return node.thickness;
        default:
            throw new core_1.UnreachableCaseError(node);
    }
};
exports.vsize = vsize;
var add = function (a, b) { return a + b; };
var zero = 0;
var sum = function (values) { return values.reduce(add, zero); };
var max = function (values) { return Math.max.apply(Math, __spreadArray([], __read(values), false)); };
var hlistWidth = function (nodes) {
    return sum(nodes.map(exports.getWidth));
};
exports.hlistWidth = hlistWidth;
var hlistHeight = function (nodes) {
    return max(nodes.map(exports.getHeight));
};
var hlistDepth = function (nodes) { return max(nodes.map(exports.getDepth)); };
var vlistVsize = function (nodes) { return sum(nodes.map(exports.vsize)); };
var makeStaticHBox = function (nodes, context) {
    var dim = {
        width: (0, exports.hlistWidth)(nodes),
        height: hlistHeight(nodes),
        depth: hlistDepth(nodes),
    };
    var content = {
        type: "static",
        nodes: nodes,
    };
    return makeHBox(dim, content, context);
};
exports.makeStaticHBox = makeStaticHBox;
var makeCursorHBox = function (left, right, context) {
    var dim = {
        width: (0, exports.hlistWidth)(left) + (0, exports.hlistWidth)(right),
        height: Math.max(hlistHeight(left), hlistHeight(right)),
        depth: Math.max(hlistDepth(left), hlistDepth(right)),
    };
    var content = {
        type: "cursor",
        left: left,
        right: right,
    };
    return makeHBox(dim, content, context);
};
exports.makeCursorHBox = makeCursorHBox;
var makeSelectionHBox = function (left, selection, right, context) {
    var dim = {
        width: (0, exports.hlistWidth)(left) + (0, exports.hlistWidth)(selection) + (0, exports.hlistWidth)(right),
        height: Math.max(hlistHeight(left), hlistHeight(selection), hlistHeight(right)),
        depth: Math.max(hlistDepth(left), hlistDepth(selection), hlistDepth(right)),
    };
    var content = {
        type: "selection",
        left: left,
        selection: selection,
        right: right,
    };
    return makeHBox(dim, content, context);
};
exports.makeSelectionHBox = makeSelectionHBox;
var makeVBox = function (width, node, upList, dnList, context) {
    var dim = {
        width: width,
        depth: dnList.length > 0
            ? vlistVsize(dnList) + (0, exports.getDepth)(node)
            : (0, exports.getDepth)(node),
        height: upList.length > 0
            ? vlistVsize(upList) + (0, exports.getHeight)(node)
            : (0, exports.getHeight)(node),
    };
    var upListCopy = __spreadArray([], __read(upList), false);
    // TODO: get rid of the need to reverse the uplist
    var nodeList = __spreadArray(__spreadArray(__spreadArray([], __read(upListCopy.reverse()), false), [node], false), __read(dnList), false);
    return __assign(__assign({ type: "VBox" }, dim), { shift: 0, content: nodeList, fontSize: (0, utils_1.fontSizeForContext)(context), style: {} });
};
exports.makeVBox = makeVBox;
var getConstantValue = function (constant, context) {
    var font = context.fontData.font;
    var fontSize = (0, utils_1.fontSizeForContext)(context);
    return (constant.value * fontSize) / font.head.unitsPerEm;
};
exports.getConstantValue = getConstantValue;
