import * as Layout from "./layout";
import type { FontData } from "@math-blocks/opentype";
declare type Style = {
    readonly fill?: string;
    readonly stroke?: string;
};
declare type Common = {
    readonly id?: number;
    readonly style: Style;
};
export declare type Group = {
    readonly type: "group";
    readonly x: number;
    readonly y: number;
    readonly bounds: Layout.Dim;
    readonly children: readonly Node[];
} & Common;
export declare type Glyph = {
    readonly type: "char";
    readonly x: number;
    readonly y: number;
    readonly width: number;
    readonly glyph: Layout.Glyph;
} & Common;
export declare type Line = {
    readonly type: "line";
    readonly x1: number;
    readonly y1: number;
    readonly x2: number;
    readonly y2: number;
    readonly thickness: number;
} & Common;
export declare type Rect = {
    readonly type: "rect";
    readonly x: number;
    readonly y: number;
    readonly width: number;
    readonly height: number;
    readonly fill?: string;
    readonly stroke?: string;
    readonly flag?: "start" | "end";
} & Common;
export declare type Node = Group | Glyph | Line | Rect;
export declare type Point = {
    readonly x: number;
    readonly y: number;
};
export declare type LayoutCursor = {
    readonly parent: number;
    readonly prev: number;
    readonly next: number;
    readonly selection: boolean;
};
declare type Options = {
    readonly showCursor?: boolean;
    readonly debug?: boolean;
};
export declare type Scene = {
    readonly width: number;
    readonly height: number;
    readonly content: Group;
    readonly selection: Group;
    readonly hitboxes: Group;
    readonly debug: Group;
};
export declare const processBox: (box: Layout.HBox, fontData: FontData, options?: Options) => Scene;
declare type Side = "left" | "right";
declare type Intersection = {
    readonly type: "content";
    readonly id: number;
    readonly side: Side;
} | {
    readonly type: "padding";
    readonly flag: "start" | "end";
};
export declare const findIntersections: (point: Point, node: Group, translation?: Point | undefined) => Intersection[];
export {};
