"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findIntersections = exports.processBox = void 0;
var core_1 = require("@math-blocks/core");
var Layout = __importStar(require("./layout"));
var processHRule = function (hrule, loc) {
    var advance = Layout.getWidth(hrule);
    return {
        type: "line",
        x1: loc.x,
        y1: loc.y,
        x2: loc.x + advance,
        y2: loc.y,
        thickness: hrule.thickness,
        style: {
            stroke: hrule.style.color,
        },
        id: hrule.id,
    };
};
var processGlyph = function (glyph, loc) {
    return {
        type: "char",
        x: loc.x,
        y: loc.y,
        width: Layout.getWidth(glyph),
        glyph: glyph,
        style: {
            fill: glyph.style.color,
        },
        id: glyph.id,
    };
};
var CURSOR_WIDTH = 2;
var processHBox = function (box, loc, context) {
    var e_1, _a;
    var pen = { x: 0, y: 0 };
    var selectionBoxes = [];
    var children = [];
    var hasSelection = !context.inSelection && box.content.type === "selection";
    var layer = context.layer, font = context.fontData.font;
    var fontSize = box.fontSize;
    var parenMetrics = font.getGlyphMetrics(font.getGlyphID(")"));
    // This assumes that parenMetrics.height < font.head.unitsPerEm
    var overshoot = (font.head.unitsPerEm - parenMetrics.height) / 2;
    var ascent = ((parenMetrics.bearingY + overshoot) * fontSize) / font.head.unitsPerEm;
    var debugStyle = {
        fill: "none",
        stroke: "red",
    };
    var cancelRegion = null;
    var sections = [];
    if (box.content.type === "static") {
        sections.push(box.content.nodes);
    }
    else if (box.content.type === "cursor") {
        sections.push(box.content.left);
        sections.push(box.content.right);
    }
    else {
        sections.push(box.content.left);
        sections.push(box.content.selection);
        sections.push(box.content.right);
    }
    sections.forEach(function (section, index) {
        var isSelection = hasSelection && index === 1;
        // There should only be two sections max.  If there are two sections
        // then we should draw a cursor in between the two of them.
        if (index === 1 &&
            !hasSelection &&
            !context.inSelection &&
            context.showCursor &&
            layer === "selection") {
            // Draw the cursor.
            children.push({
                type: "rect",
                x: pen.x - CURSOR_WIDTH / 2,
                y: pen.y - ascent,
                width: CURSOR_WIDTH,
                height: fontSize,
                style: {},
            });
        }
        section.forEach(function (node) {
            // We've encountered a new cancel id or no cancel id.  Finalize the
            // current cancel region and reset for the next one.
            if (cancelRegion &&
                cancelRegion.id !== node.style.cancel &&
                layer === "content") {
                children.push({
                    type: "line",
                    x1: cancelRegion.xMin,
                    y1: cancelRegion.yMax,
                    x2: cancelRegion.xMax,
                    y2: cancelRegion.yMin,
                    thickness: 5,
                    style: {},
                });
                cancelRegion = null;
            }
            // If the node has a cancel id, continue the current cancel region
            // or start a new one if there isn't one.
            if (typeof node.style.cancel === "number" && layer == "content") {
                var yMin = -Math.max(Layout.getHeight(node), ascent);
                var height_1 = Math.max(Layout.getHeight(node) + Layout.getDepth(node), fontSize);
                var yMax = yMin + height_1;
                if (!cancelRegion) {
                    cancelRegion = {
                        id: node.style.cancel,
                        xMin: pen.x,
                        xMax: pen.x,
                        yMin: yMin,
                        yMax: yMax,
                    };
                }
                else {
                    cancelRegion.yMin = Math.min(cancelRegion.yMin, yMin);
                    cancelRegion.yMax = Math.max(cancelRegion.yMax, yMax);
                }
            }
            if (isSelection && layer === "selection") {
                var yMin = -Math.max(Layout.getHeight(node), ascent);
                var height_2 = Math.max(Layout.getHeight(node) + Layout.getDepth(node), fontSize);
                // Draw the cursor.
                selectionBoxes.push({
                    type: "rect",
                    x: pen.x,
                    y: yMin,
                    width: Layout.getWidth(node),
                    height: height_2,
                    style: {},
                });
            }
            var advance = Layout.getWidth(node);
            var height = Layout.getHeight(node);
            var depth = Layout.getDepth(node);
            switch (node.type) {
                case "HBox":
                case "VBox": {
                    var child = _processBox(node, { x: pen.x, y: pen.y + node.shift }, __assign(__assign({}, context), { inSelection: context.inSelection || hasSelection }));
                    // We always have to include child groups regardless of the
                    // layer.  TODO: drop this once we flatten the scene graph.
                    children.push(child);
                    if (layer === "debug") {
                        children.push({
                            type: "rect",
                            id: node.id,
                            x: pen.x,
                            y: pen.y - height,
                            width: advance,
                            height: depth + height,
                            style: debugStyle,
                        });
                    }
                    break;
                }
                case "HRule": {
                    var child = processHRule(node, pen);
                    if (layer === "content") {
                        children.push(child);
                    }
                    break;
                }
                case "Glyph": {
                    var child = processGlyph(node, pen);
                    if (layer === "content") {
                        children.push(child);
                    }
                    if (layer === "debug") {
                        children.push({
                            type: "rect",
                            id: node.id,
                            x: pen.x,
                            y: pen.y - height,
                            width: advance,
                            height: depth + height,
                            style: debugStyle,
                        });
                    }
                    if (layer === "hitboxes") {
                        // TODO: do a second pass on the hitboxes to expand them
                        // to their full height
                        children.push({
                            type: "rect",
                            id: node.id,
                            x: pen.x,
                            y: pen.y - height,
                            width: advance,
                            height: depth + height,
                            style: debugStyle,
                        });
                    }
                    break;
                }
                case "Kern":
                    if (node.flag) {
                        if (layer === "hitboxes") {
                            // TODO: do a second pass on the hitboxes to expand
                            // them to their full height
                            children.push({
                                type: "rect",
                                flag: node.flag,
                                x: pen.x,
                                y: pen.y - box.height,
                                width: node.size,
                                height: box.depth + box.height,
                                style: debugStyle,
                            });
                        }
                    }
                    // We don't need to include kerns in the output since we include
                    // the cursor or select rectangle in the scene graph.
                    break;
                default:
                    throw new core_1.UnreachableCaseError(node);
            }
            pen.x += advance;
            if (node.style.cancel && cancelRegion) {
                cancelRegion.xMax = pen.x;
            }
        });
    });
    if (cancelRegion !== null && layer === "content") {
        var thickness = (fontSize * font.math.constants.fractionRuleThickness.value) /
            font.head.unitsPerEm;
        children.push({
            type: "line",
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            x1: cancelRegion.xMin,
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            y1: cancelRegion.yMax,
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            x2: cancelRegion.xMax,
            // @ts-expect-error: TypeScript doesn't understand forEach loops
            y2: cancelRegion.yMin,
            thickness: thickness,
            style: {},
        });
        cancelRegion = null;
    }
    // Draw the selection.
    if (layer === "selection") {
        try {
            for (var selectionBoxes_1 = __values(selectionBoxes), selectionBoxes_1_1 = selectionBoxes_1.next(); !selectionBoxes_1_1.done; selectionBoxes_1_1 = selectionBoxes_1.next()) {
                var selectionBox = selectionBoxes_1_1.value;
                children.unshift(__assign(__assign({}, selectionBox), { fill: "Highlight", stroke: "none" }));
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (selectionBoxes_1_1 && !selectionBoxes_1_1.done && (_a = selectionBoxes_1.return)) _a.call(selectionBoxes_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    }
    return {
        type: "group",
        x: loc.x,
        y: loc.y,
        bounds: box,
        children: children,
        style: layer === "content"
            ? {
                fill: box.style.color,
                stroke: box.style.color,
            }
            : {},
        id: box.id,
    };
};
var processVBox = function (box, loc, context) {
    var pen = { x: 0, y: 0 };
    pen.y -= box.height;
    var children = [];
    var layer = context.layer;
    var debugStyle = {
        fill: "none",
        stroke: "red",
    };
    box.content.forEach(function (node) {
        var width = Layout.getWidth(node);
        var height = Layout.getHeight(node);
        var depth = Layout.getDepth(node);
        switch (node.type) {
            case "HBox":
            case "VBox": {
                // TODO: reconsider whether we should be taking the shift into
                // account when computing the height, maybe we can drop this
                // and simplify things.  The reason why we zero out the shift
                // here is that when we render a box inside of a vbox, the shift
                // is a horizontal shift as opposed to a vertical one.
                // I'm not sure we can do this properly since how the shift is
                // used depends on the parent box type.  We could pass that info
                // to the getHeight() function... we should probably do an audit
                // of all the callsites for getHeight()
                var height_3 = Layout.getHeight(__assign(__assign({}, node), { shift: 0 }));
                var depth_1 = Layout.getDepth(__assign(__assign({}, node), { shift: 0 }));
                pen.y += height_3;
                // TODO: see if we can get rid of this check in the future
                if (Number.isNaN(pen.y)) {
                    // eslint-disable-next-line no-debugger
                    debugger;
                }
                var child = _processBox(node, { x: pen.x + node.shift, y: pen.y }, context);
                // We always have to include child groups regardless of the
                // layer.  TODO: drop this once we flatten the scene graph.
                children.push(child);
                if (layer === "debug") {
                    children.push({
                        type: "rect",
                        id: node.id,
                        x: pen.x + node.shift,
                        y: pen.y - height_3,
                        width: width,
                        height: depth_1 + height_3,
                        style: debugStyle,
                    });
                }
                pen.y += depth_1;
                break;
            }
            case "HRule": {
                pen.y += height;
                var child = processHRule(node, pen);
                if (layer === "content") {
                    children.push(child);
                }
                pen.y += depth;
                break;
            }
            case "Glyph": {
                // Although there currently isn't anything that uses a glyph
                // in a vbox, we'll likely need it for accents.
                pen.y += height;
                var child = processGlyph(node, pen);
                if (layer === "content") {
                    children.push(child);
                }
                if (layer === "debug") {
                    children.push({
                        type: "rect",
                        id: node.id,
                        x: pen.x,
                        y: pen.y,
                        width: width,
                        height: depth + height,
                        style: debugStyle,
                    });
                }
                if (layer === "hitboxes") {
                    // TODO: do a second pass on the hitboxes to expand them
                    // to their full height
                    children.push({
                        type: "rect",
                        id: node.id,
                        x: pen.x,
                        y: pen.y,
                        width: width,
                        height: depth + height,
                        style: debugStyle,
                    });
                }
                pen.y += depth;
                break;
            }
            case "Kern":
                pen.y += node.size;
                break;
            default:
                throw new core_1.UnreachableCaseError(node);
        }
    });
    return {
        type: "group",
        x: loc.x,
        y: loc.y,
        bounds: box,
        children: children,
        style: layer === "content"
            ? {
                fill: box.style.color,
                stroke: box.style.color,
            }
            : {},
        id: box.id,
    };
};
var _processBox = function (box, loc, context) {
    switch (box.type) {
        case "HBox":
            return processHBox(box, loc, context);
        case "VBox":
            return processVBox(box, loc, context);
    }
};
var processBox = function (box, fontData, options) {
    if (options === void 0) { options = {}; }
    var loc = { x: 0, y: Layout.getHeight(box) };
    var context = {
        showCursor: !!options.showCursor,
        inSelection: false,
        fontData: fontData,
        layer: "content",
    };
    var contentLayer = _processBox(box, loc, context);
    var selectionLayer = _processBox(box, loc, __assign(__assign({}, context), { layer: "selection" }));
    var debugLayer = _processBox(box, loc, __assign(__assign({}, context), { layer: "debug" }));
    var hitboxes = _processBox(box, loc, __assign(__assign({}, context), { layer: "hitboxes" }));
    var scene = {
        width: contentLayer.bounds.width,
        height: contentLayer.bounds.height + contentLayer.bounds.depth,
        content: contentLayer,
        selection: selectionLayer,
        debug: debugLayer,
        hitboxes: hitboxes,
    };
    return scene;
};
exports.processBox = processBox;
var isPointInBounds = function (point, bounds) {
    if (point.x > bounds.x &&
        point.x < bounds.x + bounds.width / 2 &&
        point.y > bounds.y &&
        point.y < bounds.y + bounds.height) {
        return "left";
    }
    if (point.x > bounds.x + bounds.width / 2 &&
        point.x < bounds.x + bounds.width &&
        point.y > bounds.y &&
        point.y < bounds.y + bounds.height) {
        return "right";
    }
    return undefined;
};
var getBounds = function (child, translation) {
    return child.type === "group"
        ? {
            x: child.x + translation.x,
            y: child.y + translation.y - child.bounds.height,
            width: child.bounds.width,
            height: child.bounds.height + child.bounds.depth,
        }
        : {
            x: child.x + translation.x,
            y: child.y + translation.y,
            width: child.width,
            height: child.height,
        };
};
var getIntersection = function (child, side) {
    if (child.id) {
        return {
            type: "content",
            id: child.id,
            side: side,
        };
    }
    else if (child.type === "rect" && child.flag) {
        return {
            type: "padding",
            flag: child.flag,
        };
    }
};
var findIntersections = function (point, node, // must be the group containing the debug bounding rectangles
translation) {
    var e_2, _a, e_3, _b, e_4, _c;
    var result = [];
    translation = translation || {
        x: node.x,
        y: node.y,
    };
    try {
        for (var _d = __values(node.children), _e = _d.next(); !_e.done; _e = _d.next()) {
            var child = _e.value;
            if (child.type !== "group" && child.type !== "rect") {
                throw new Error("Unexpected node type in hitboxes");
            }
            var bounds = getBounds(child, translation);
            var side = isPointInBounds(point, bounds);
            if (side) {
                var intersection = getIntersection(child, side);
                if (intersection) {
                    result.push(intersection);
                }
                if (child.type === "group") {
                    var newTranslation = {
                        x: child.x + translation.x,
                        y: child.y + translation.y,
                    };
                    result.push.apply(result, __spreadArray([], __read((0, exports.findIntersections)(point, child, newTranslation)), false));
                }
            }
        }
    }
    catch (e_2_1) { e_2 = { error: e_2_1 }; }
    finally {
        try {
            if (_e && !_e.done && (_a = _d.return)) _a.call(_d);
        }
        finally { if (e_2) throw e_2.error; }
    }
    if (result.length === 0) {
        // iterate through all of the children again and find any bounds that
        // intersect with the vertical line crossing with `point` on it
        var candidateIntersections = [];
        try {
            for (var _f = __values(node.children), _g = _f.next(); !_g.done; _g = _f.next()) {
                var child = _g.value;
                if (child.type !== "group" && child.type !== "rect") {
                    throw new Error("Unexpected node type in hitboxes");
                }
                var bounds_1 = getBounds(child, translation);
                // We decide which side of the child to put the cursor on once we
                // decide on which candidate is the closest vertically to the point.
                if (point.x > bounds_1.x && point.x < bounds_1.x + bounds_1.width) {
                    candidateIntersections.push(child);
                }
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (_g && !_g.done && (_b = _f.return)) _b.call(_f);
            }
            finally { if (e_3) throw e_3.error; }
        }
        // If there are no candidates don't bother finding the closest one.
        if (candidateIntersections.length === 0) {
            return result;
        }
        var _h = __read(candidateIntersections), int = _h[0], rest = _h.slice(1);
        // The closest candidate is the one with the smallest vertical distance
        // from the point.
        var bounds = getBounds(int, translation);
        var minDist = Math.min(Math.abs(bounds.y - point.y), Math.abs(bounds.y + bounds.height - point.y));
        var closest = int;
        try {
            for (var rest_1 = __values(rest), rest_1_1 = rest_1.next(); !rest_1_1.done; rest_1_1 = rest_1.next()) {
                var candidate = rest_1_1.value;
                bounds = getBounds(candidate, translation);
                var dist = Math.min(Math.abs(bounds.y - point.y), Math.abs(bounds.y + bounds.height - point.y));
                if (dist < minDist) {
                    minDist = dist;
                    closest = candidate;
                }
            }
        }
        catch (e_4_1) { e_4 = { error: e_4_1 }; }
        finally {
            try {
                if (rest_1_1 && !rest_1_1.done && (_c = rest_1.return)) _c.call(rest_1);
            }
            finally { if (e_4) throw e_4.error; }
        }
        // Determine whether the cursor should be to the left or the right of
        // the closest node.
        bounds = getBounds(closest, translation);
        var side = point.x < bounds.x + bounds.width / 2 ? "left" : "right";
        if (side) {
            var intersection = getIntersection(closest, side);
            if (intersection) {
                result.push(intersection);
            }
            if (closest.type === "group") {
                var newTranslation = {
                    x: closest.x + translation.x,
                    y: closest.y + translation.y,
                };
                result.push.apply(result, __spreadArray([], __read((0, exports.findIntersections)(point, closest, newTranslation)), false));
            }
        }
    }
    return result;
};
exports.findIntersections = findIntersections;
