"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typeset = exports.typesetZipper = void 0;
var core_1 = require("@math-blocks/core");
var Editor = __importStar(require("@math-blocks/editor"));
var Layout = __importStar(require("./layout"));
var scene_graph_1 = require("./scene-graph");
var enums_1 = require("./enums");
var utils_1 = require("./utils");
var delimited_1 = require("./typesetters/delimited");
var frac_1 = require("./typesetters/frac");
var limits_1 = require("./typesetters/limits");
var root_1 = require("./typesetters/root");
var subsup_1 = require("./typesetters/subsup");
var table_1 = require("./typesetters/table");
var atom_1 = require("./typesetters/atom");
var NodeType = Editor.NodeType;
var typesetRow = function (row, context, padFirstOperator) {
    var box = Layout.makeStaticHBox(typesetNodes(row.children, context, undefined, undefined, padFirstOperator), context);
    box.id = row.id;
    box.style = __assign(__assign({}, box.style), { color: row.style.color });
    if (context.renderMode === enums_1.RenderMode.Dynamic) {
        ensureMinDepthAndHeight(box, context);
    }
    return box;
};
/**
 * This function is used to guarantee a certain depth/height for a box that is
 * equal to the depth/height of the cursor or selection rectangle.  This is
 * useful for typesetting while editing since it minimize changes to vertical
 * size of the rendered math.
 *
 * WARNING: This function mutates `box`.
 *
 * TODO: add originalDepth and originalHeight so that getDelimiter can make its
 * decisions based on the original dimensions of the box.
 *
 * @param {Mutable<Layout.Dim>} dim
 * @param {Context} context
 * @return {void}
 */
var ensureMinDepthAndHeight = function (dim, context) {
    var font = context.fontData.font;
    var fontSize = (0, utils_1.fontSizeForContext)(context);
    var parenMetrics = font.getGlyphMetrics(font.getGlyphID(")"));
    // This assumes that parenMetrics.height < font.head.unitsPerEm
    var overshoot = (font.head.unitsPerEm - parenMetrics.height) / 2;
    var depth = ((parenMetrics.height - parenMetrics.bearingY + overshoot) * fontSize) /
        font.head.unitsPerEm;
    dim.depth = Math.max(dim.depth, depth);
    var height = ((parenMetrics.bearingY + overshoot) * fontSize) / font.head.unitsPerEm;
    dim.height = Math.max(dim.height, height);
};
var getTypesetChildFromZipper = function (zipper, focus) {
    return function (index, context, padFirstOperator) {
        if (index < focus.left.length) {
            var child = focus.left[index];
            return child && typesetRow(child, context, padFirstOperator);
        }
        else if (index === focus.left.length) {
            return _typesetZipper(zipper, context, padFirstOperator);
        }
        else {
            var child = focus.right[index - focus.left.length - 1];
            return child && typesetRow(child, context, padFirstOperator);
        }
    };
};
var getTypesetChildFromNodes = function (children) {
    return function (index, context, padFirstOperator) {
        var child = children[index];
        return child && typesetRow(child, context, padFirstOperator);
    };
};
var typesetFocus = function (focus, zipper, context, prevEditNode, prevLayoutNode) {
    var typesetChild = getTypesetChildFromZipper(zipper, focus);
    switch (focus.type) {
        case "zfrac": {
            return (0, frac_1.typesetFrac)(typesetChild, focus, context);
        }
        case "zsubsup": {
            return (0, subsup_1.typesetSubsup)(typesetChild, focus, context, prevEditNode, prevLayoutNode);
        }
        case "zroot": {
            return (0, root_1.typesetRoot)(typesetChild, focus, context);
        }
        case "zlimits": {
            return (0, limits_1.typesetLimits)(typesetChild, focus, context, typesetNode);
        }
        case "zdelimited": {
            return (0, delimited_1.typesetDelimited)(typesetChild, focus, context);
        }
        case "ztable": {
            return (0, table_1.typesetTable)(typesetChild, focus, context, zipper);
        }
        default:
            throw new core_1.UnreachableCaseError(focus);
    }
};
var typesetNode = function (node, context, prevEditNode, prevLayoutNode, padFirstOperator) {
    switch (node.type) {
        case NodeType.Row: {
            // The only time this can happen is if limits.inner is a row
            return typesetRow(node, context);
        }
        case NodeType.Frac: {
            var typesetChild = getTypesetChildFromNodes(node.children);
            return (0, frac_1.typesetFrac)(typesetChild, node, context);
        }
        case NodeType.SubSup: {
            return (0, subsup_1.typesetSubsup)(getTypesetChildFromNodes(node.children), node, context, prevEditNode, prevLayoutNode);
        }
        case NodeType.Root: {
            var typesetChild = getTypesetChildFromNodes(node.children);
            return (0, root_1.typesetRoot)(typesetChild, node, context);
        }
        case NodeType.Limits: {
            var typesetChild = getTypesetChildFromNodes(node.children);
            return (0, limits_1.typesetLimits)(typesetChild, node, context, typesetNode);
        }
        case NodeType.Delimited: {
            var typesetChild = getTypesetChildFromNodes(node.children);
            return (0, delimited_1.typesetDelimited)(typesetChild, node, context);
        }
        case NodeType.Table: {
            var typesetChild = getTypesetChildFromNodes(node.children);
            return (0, table_1.typesetTable)(typesetChild, node, context);
        }
        case "char": {
            return (0, atom_1.maybeAddOperatorPadding)(prevEditNode, node, context, padFirstOperator);
        }
        default:
            throw new core_1.UnreachableCaseError(node);
    }
};
var typesetNodes = function (nodes, context, prevChild, prevLayoutNode, padFirstOperator) {
    return nodes.map(function (child, index) {
        var result = typesetNode(child, context, prevChild, prevLayoutNode, index === 0 ? padFirstOperator : undefined);
        prevLayoutNode = result;
        prevChild = child;
        return result;
    });
};
var _typesetZipper = function (zipper, context, padFirstOperator) {
    // The bottommost crumb is the outermost row
    var _a = __read(zipper.breadcrumbs), crumb = _a[0], restCrumbs = _a.slice(1);
    if (crumb) {
        var row = crumb.row;
        var nextZipper = __assign(__assign({}, zipper), { breadcrumbs: restCrumbs });
        var nodes = [];
        nodes.push.apply(nodes, __spreadArray([], __read(typesetNodes(row.left, context, undefined, undefined, padFirstOperator)), false));
        nodes.push(typesetFocus(crumb.focus, nextZipper, context, row.left[row.left.length - 1], // previous edit node
        nodes[nodes.length - 1]));
        nodes.push.apply(nodes, __spreadArray([], __read(typesetNodes(row.right, context, crumb.focus, // previous edit node
        nodes[nodes.length - 1], // previous layout node
        padFirstOperator)), false));
        var box = Layout.makeStaticHBox(nodes, context);
        box.id = row.id;
        box.style = __assign(__assign({}, box.style), { color: row.style.color });
        if (context.renderMode === enums_1.RenderMode.Dynamic) {
            ensureMinDepthAndHeight(box, context);
        }
        return box;
    }
    else {
        var row = zipper.row;
        var input = __spreadArray(__spreadArray(__spreadArray([], __read(row.left), false), __read(row.selection), false), __read(row.right), false);
        var output = typesetNodes(input, context, undefined, undefined, padFirstOperator);
        var firstCut = row.left.length;
        var secondCut = firstCut + row.selection.length;
        var left = output.slice(0, firstCut);
        var selection = output.slice(firstCut, secondCut);
        var right = output.slice(secondCut);
        var box = (selection.length > 0
            ? Layout.makeSelectionHBox(left, selection, right, context)
            : Layout.makeCursorHBox(left, right, context));
        box.id = row.id;
        box.style = __assign(__assign({}, box.style), { color: row.style.color });
        if (context.renderMode === enums_1.RenderMode.Dynamic) {
            ensureMinDepthAndHeight(box, context);
        }
        return box;
    }
};
var typesetZipper = function (zipper, context, options) {
    if (options === void 0) { options = {}; }
    var box = _typesetZipper(zipper, context);
    return (0, scene_graph_1.processBox)(box, context.fontData, options);
};
exports.typesetZipper = typesetZipper;
var typeset = function (node, context, options) {
    if (options === void 0) { options = {}; }
    var box = typesetNode(node, context);
    return (0, scene_graph_1.processBox)(box, context.fontData, options);
};
exports.typeset = typeset;
