"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typesetAtom = exports.maybeAddOperatorPadding = void 0;
var Editor = __importStar(require("@math-blocks/editor"));
var Layout = __importStar(require("../layout"));
var utils_1 = require("../utils");
var canBeUnary = function (char) {
    var unaryOperators = [
        "+",
        "\u2212",
        "\u00B1", // \pm
    ];
    return unaryOperators.includes(char);
};
var shouldHavePadding = function (prevNode, currentNode, context) {
    var currentChar = currentNode.value;
    // We only add padding around operators, so if we get a non-operator char
    // we can return early.
    if (!Editor.util.isOperator(currentNode)) {
        return false;
    }
    // No operators get padding when `cramped` is true
    if (context.cramped) {
        return false;
    }
    // If the currentChar can be unary we check a number of situations where it
    // should be unary and don't give it any padding in those situations.
    if (canBeUnary(currentChar)) {
        if (!prevNode ||
            (prevNode.type === "char" && Editor.util.isOperator(prevNode)) ||
            prevNode.type === "limits" ||
            prevNode.type === "zlimits") {
            return false;
        }
    }
    // All other operators should have padding around them.
    return true;
};
var maybeAddOperatorPadding = function (prevNode, currentNode, context, padOperator) {
    var glyph = (0, exports.typesetAtom)(currentNode, context);
    var fontSize = (0, utils_1.fontSizeForContext)(context);
    var result = (padOperator && Editor.util.isOperator(currentNode)) ||
        shouldHavePadding(prevNode, currentNode, context)
        ? Layout.makeStaticHBox([
            Layout.makeKern(fontSize / 4),
            glyph,
            Layout.makeKern(fontSize / 4),
        ], context)
        : glyph;
    if (result !== glyph) {
        result.id = glyph.id;
        delete glyph.id;
        // Move the style to the result so that cancel overlays are
        // continuous even when they include an operator with padding.
        result.style = glyph.style;
        glyph.style = {};
    }
    return result;
};
exports.maybeAddOperatorPadding = maybeAddOperatorPadding;
var typesetAtom = function (node, context) {
    var font = context.fontData.font;
    var glyphID = font.getGlyphID(node.value);
    var glyph = Layout.makeGlyph(node.value, glyphID, context);
    // Convert individual glyphs to italic glyphs if they exist in the
    // current font.
    if (/[a-z]/.test(node.value) && !context.operator) {
        var offset = node.value.charCodeAt(0) - "a".charCodeAt(0);
        var char = String.fromCodePoint(0x1d44e + offset);
        var glyphID_1 = font.getGlyphID(char);
        glyph = Layout.makeGlyph(char, glyphID_1, context);
    }
    glyph.id = node.id;
    glyph.style = node.style;
    glyph.pending = node.pending;
    return glyph;
};
exports.typesetAtom = typesetAtom;
