"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typesetFrac = void 0;
var Layout = __importStar(require("../layout"));
var utils_1 = require("../utils");
var enums_1 = require("../enums");
var makeList = function (size, box) { return [Layout.makeKern(size), box]; };
var childContextForFrac = function (context) {
    var _a;
    var mathStyle = context.mathStyle;
    var childMathStyle = (_a = {},
        _a[enums_1.MathStyle.Display] = enums_1.MathStyle.Text,
        _a[enums_1.MathStyle.Text] = enums_1.MathStyle.Script,
        _a[enums_1.MathStyle.Script] = enums_1.MathStyle.ScriptScript,
        _a[enums_1.MathStyle.ScriptScript] = enums_1.MathStyle.ScriptScript,
        _a)[mathStyle];
    var childContext = __assign(__assign({}, context), { mathStyle: childMathStyle });
    return childContext;
};
var typesetFrac = function (typesetChild, node, context) {
    var childContext = childContextForFrac(context);
    var numBox = typesetChild(0, childContext);
    var denBox = typesetChild(1, childContext);
    if (!numBox || !denBox) {
        throw new Error("The numerator and denominator must both be defined");
    }
    var mathStyle = context.mathStyle;
    var constants = context.fontData.font.math.constants;
    var fontSize = (0, utils_1.fontSizeForContext)(context);
    var thickness = (fontSize * constants.fractionRuleThickness.value) / 1000;
    var shift = (fontSize * constants.axisHeight.value) / 1000;
    // If useDisplayStyle is false then we need to reduce the font size of
    // numerators and denominators
    var useDisplayStyle = mathStyle === enums_1.MathStyle.Display;
    var minDenGap = useDisplayStyle
        ? (fontSize * constants.fractionDenomDisplayStyleGapMin.value) / 1000
        : (fontSize * constants.fractionDenominatorGapMin.value) / 1000;
    var minNumGap = useDisplayStyle
        ? (fontSize * constants.fractionNumDisplayStyleGapMin.value) / 1000
        : (fontSize * constants.fractionNumeratorGapMin.value) / 1000;
    // TODO: fix this code once we have debugging outlines in place, right now
    // gap between the fraction bar and the numerator/denominator is too big
    // when the font has been scaled.
    // Check context.renderMode === RenderMode.Static when doing so.
    // const numeratorShift = useDisplayStyle
    //     ? (fontSize * constants.fractionNumeratorDisplayStyleShiftUp) / 1000
    //     : (fontSize * constants.fractionNumeratorShiftUp) / 1000;
    // const denominatorShift = useDisplayStyle
    //     ? (fontSize * constants.fractionDenominatorDisplayStyleShiftDown) / 1000
    //     : (fontSize * constants.fractionDenominatorShiftDown) / 1000;
    // const numGap = Math.max(numeratorShift - numBox.depth - shift, minNumGap);
    // const denGap = Math.max(
    //     shift + denominatorShift - denBox.height,
    //     minDenGap,
    // );
    var multiplier = (0, utils_1.multiplierForContext)(context);
    var endPadding = thickness; // add extra space around the numerator and denominator
    var width = Math.max(Math.max(Layout.getWidth(numBox), Layout.getWidth(denBox)), // TODO: calculate this based on current font size
    30 * multiplier) +
        2 * endPadding;
    // center the numerator
    if (Layout.getWidth(numBox) < width) {
        var kernSize = (width - Layout.getWidth(numBox)) / 2;
        numBox = Layout.rebox(numBox, Layout.makeKern(kernSize, "start"), Layout.makeKern(kernSize, "end"));
    }
    // center the denominator
    if (Layout.getWidth(denBox) < width) {
        var kernSize = (width - Layout.getWidth(denBox)) / 2;
        denBox = Layout.rebox(denBox, Layout.makeKern(kernSize, "start"), Layout.makeKern(kernSize, "end"));
    }
    var upList = makeList(minNumGap, numBox);
    var dnList = makeList(minDenGap, denBox);
    var stroke = Layout.makeStaticHBox([Layout.makeHRule(thickness, width)], context);
    var fracBox = Layout.makeVBox(width, stroke, upList, dnList, context);
    fracBox.shift = -shift;
    fracBox.id = node.id;
    fracBox.style = node.style;
    return fracBox;
};
exports.typesetFrac = typesetFrac;
