"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typesetRoot = void 0;
var Layout = __importStar(require("../layout"));
var enums_1 = require("../enums");
var enums_2 = require("../enums");
var utils_1 = require("../utils");
var typesetRoot = function (typesetChild, node, context) {
    var _a;
    var degreeContext = __assign(__assign({}, context), { 
        // It doesn't matter what the mathStyle is of the parent, we
        // always use ScriptScript for root indicies.
        mathStyle: enums_1.MathStyle.ScriptScript });
    var degree = typesetChild(0, degreeContext);
    var radicand = typesetChild(1, context);
    if (!radicand) {
        throw new Error("Radicand must be defined");
    }
    var multiplier = (0, utils_1.multiplierForContext)(context);
    // Give the radicand a minimal width in case it's empty
    radicand.width = Math.max(radicand.width, 30 * multiplier);
    var thresholdOptions = {
        value: "sum",
        strict: true,
    };
    var surd = Layout.makeStaticHBox([(0, utils_1.makeDelimiter)("\u221A", radicand, thresholdOptions, context)], context);
    var fontSize = (0, utils_1.fontSizeForContext)(context);
    var font = context.fontData.font;
    var constants = context.fontData.font.math.constants;
    var thickness = (fontSize * constants.radicalRuleThickness.value) /
        font.head.unitsPerEm;
    var endPadding = thickness; // Add extra space at the end of the radicand
    var stroke = Layout.makeHRule(thickness, radicand.width + endPadding);
    var radicalWithRule = Layout.makeVBox(radicand.width, radicand, [Layout.makeKern(6), stroke], [], context);
    // Compute the shift to align the top of the surd with the radical rule
    var shift = surd.height - radicalWithRule.height;
    surd.shift = shift;
    var root;
    if (degree) {
        var afterDegreeKern = Layout.makeKern((fontSize * constants.radicalKernAfterDegree.value) /
            font.head.unitsPerEm);
        // TODO: take into account constants.radicalKernBeforeDegree
        var beforeDegreeKern = Layout.makeKern(Math.max(0, -afterDegreeKern.size - degree.width));
        // This follows the instructions from 3.3.3.3 describing the alphabetic
        // baseline of the index.
        // https://mathml-refresh.github.io/mathml-core/#root-with-index
        //
        // NOTE: This doesn't account for an index with a large descender, but
        // neither does TeX's layout algorithm.  Most of the the time the index
        // is a single number or `n`, neither of whcih have a descender.
        var degreeBottomRaisePercent = constants.radicalDegreeBottomRaisePercent / 100;
        // We default to MathML/Word beahavior since that's what most fonts
        // seem to use.
        var algorithm = (_a = context.radicalDegreeAlgorithm) !== null && _a !== void 0 ? _a : enums_2.RadicalDegreeAlgorithm.MathML;
        switch (algorithm) {
            case enums_2.RadicalDegreeAlgorithm.OpenType:
                degree.shift =
                    shift - // match shift of surdHBox
                        // The OpenType spec says `radicalDegreeBottomRaisePercent` is
                        // with respect to the ascender of the radical glyph.
                        degreeBottomRaisePercent * surd.height;
                break;
            case enums_2.RadicalDegreeAlgorithm.MathML:
                degree.shift =
                    shift + // match shift of surdHBox
                        surd.depth - // align the index to the bottom of surdHBox
                        // The MathML Core spec says `radicalDegreeBottomRaisePercent`
                        // is with respect to the height of the radical glyph.
                        degreeBottomRaisePercent * Layout.vsize(surd);
                break;
        }
        root = Layout.makeStaticHBox([beforeDegreeKern, degree, afterDegreeKern, surd, radicalWithRule], context);
    }
    else {
        root = Layout.makeStaticHBox([surd, radicalWithRule], context);
    }
    root.width += endPadding;
    root.id = node.id;
    root.style = node.style;
    return root;
};
exports.typesetRoot = typesetRoot;
