"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typesetSubsup = void 0;
var Layout = __importStar(require("../layout"));
var enums_1 = require("../enums");
var childContextForSubsup = function (context) {
    var _a;
    var mathStyle = context.mathStyle;
    var childMathStyle = (_a = {},
        _a[enums_1.MathStyle.Display] = enums_1.MathStyle.Script,
        _a[enums_1.MathStyle.Text] = enums_1.MathStyle.Script,
        _a[enums_1.MathStyle.Script] = enums_1.MathStyle.ScriptScript,
        _a[enums_1.MathStyle.ScriptScript] = enums_1.MathStyle.ScriptScript,
        _a)[mathStyle];
    var childContext = __assign(__assign({}, context), { mathStyle: childMathStyle, cramped: true });
    return childContext;
};
var typesetSubsup = function (typesetChild, node, context, prevEditNode, prevLayoutNode) {
    var childContext = childContextForSubsup(context);
    var subBox = typesetChild(0, childContext);
    var supBox = typesetChild(1, childContext);
    if (!supBox && !subBox) {
        throw new Error("at least one of supBox and subBox must be defined");
    }
    var font = context.fontData.font;
    var width = Math.max(supBox ? Layout.getWidth(supBox) : 0, subBox ? Layout.getWidth(subBox) : 0);
    // Some atoms gets wrapped in a box to add padding to them so we need to
    // filter them out.  Anything else that's in a box is some sort of compound
    // layout structure (frac, delimited, etc.) and should have its subscript
    // and/or superscript positioned based on the size of the box.
    if ((prevEditNode === null || prevEditNode === void 0 ? void 0 : prevEditNode.type) !== "char" && prevLayoutNode) {
        var _a = font.math.constants, superscriptBaselineDropMax = _a.superscriptBaselineDropMax, subscriptBaselineDropMin = _a.subscriptBaselineDropMin;
        var baselineDropMax = Layout.getConstantValue(superscriptBaselineDropMax, context);
        var baselineDropMin = Layout.getConstantValue(subscriptBaselineDropMin, context);
        var upList_1 = [];
        var dnList_1 = [];
        if (supBox) {
            var shift = Layout.getHeight(prevLayoutNode);
            var kernShift = shift - baselineDropMax;
            upList_1.push(Layout.makeKern(kernShift));
            upList_1.push(supBox);
        }
        if (subBox) {
            var shift = Layout.getDepth(prevLayoutNode);
            var kernSize = shift - baselineDropMin;
            dnList_1.push(Layout.makeKern(kernSize));
            dnList_1.push(subBox);
        }
        var referenceNode_1 = Layout.makeKern(0); // empty reference node
        var subsupBox_1 = Layout.makeVBox(width, referenceNode_1, upList_1, dnList_1, context);
        subsupBox_1.id = node.id;
        subsupBox_1.style = node.style;
        return subsupBox_1;
    }
    var _b = font.math.constants, subscriptTopMax = _b.subscriptTopMax, superscriptBottomMin = _b.superscriptBottomMin, subscriptShiftDown = _b.subscriptShiftDown, superscriptShiftUp = _b.superscriptShiftUp, subSuperscriptGapMin = _b.subSuperscriptGapMin, superscriptBottomMaxWithSubscript = _b.superscriptBottomMaxWithSubscript;
    var bottomMin = Layout.getConstantValue(superscriptBottomMin, context);
    var shiftUp = Layout.getConstantValue(superscriptShiftUp, context);
    var topMax = Layout.getConstantValue(subscriptTopMax, context);
    var shiftDown = Layout.getConstantValue(subscriptShiftDown, context);
    var upList = [];
    if (supBox) {
        // compute shift in baseline of superscript
        var shift = Math.max(shiftUp, supBox.depth + bottomMin);
        // -supBox.depth is to align the baseline of the superscript with the
        // baseline of the base.
        // TODO: replace the up/dn list that makeVBox uses with something else
        // that doesn't require this correction
        var kernShift = -supBox.depth + shift;
        upList.push(Layout.makeKern(kernShift));
        upList.push(supBox);
    }
    var dnList = [];
    if (subBox) {
        // compute shift in baseline of subscript
        var shift = Math.max(shiftDown, subBox.height - topMax);
        if (supBox) {
            var supBoxShift = Math.max(shiftUp, supBox.depth + bottomMin);
            var supBottom = supBoxShift - supBox.depth;
            var subTop = subBox.height - shift;
            var gap = supBottom - subTop;
            var gapMin = Layout.getConstantValue(subSuperscriptGapMin, context);
            if (gap < gapMin) {
                if (upList[0].type === "Kern" && upList[1].type === "HBox") {
                    // shift the superscript up to increase the gap
                    var correction = gapMin - gap;
                    upList[0].size += correction;
                    // We can't use upList[0].size in the calculation since it
                    // includes the initial -supBox.depth to align baselines.
                    var supBottom_1 = supBoxShift + correction - upList[1].depth;
                    var supBottomMax = Layout.getConstantValue(superscriptBottomMaxWithSubscript, context);
                    if (supBottom_1 > supBottomMax) {
                        // shift both down to maintain to gap
                        var correcion = supBottom_1 - supBottomMax;
                        upList[0].size -= correcion;
                        shift += correcion; // down is positive for dnList
                    }
                }
            }
        }
        // We start with -subBox.height to align the subscript's baseline with
        // the baseline of the base it's attached to
        // TODO: replace the up/dn list that makeVBox uses with something else
        // that doesn't require this correction
        var kernSize = -subBox.height + shift;
        dnList.push(Layout.makeKern(kernSize));
        dnList.push(subBox);
    }
    var referenceNode = Layout.makeKern(0); // empty reference node
    var subsupBox = Layout.makeVBox(width, referenceNode, upList, dnList, context);
    subsupBox.id = node.id;
    subsupBox.style = node.style;
    return subsupBox;
};
exports.typesetSubsup = typesetSubsup;
