"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.typesetTable = void 0;
var Editor = __importStar(require("@math-blocks/editor"));
var Layout = __importStar(require("../layout"));
var utils_1 = require("../utils");
var enums_1 = require("../enums");
var DEFAULT_GUTTER_WIDTH = 50;
var isOperator = function (cell) {
    return (cell === null || cell === void 0 ? void 0 : cell.children.length) === 1 &&
        Editor.util.isAtom(cell.children[0], ["+", "\u2212", "=", "<", ">"]);
};
var childContextForTable = function (context) {
    var _a;
    var mathStyle = context.mathStyle;
    var childMathStyle = (_a = {},
        _a[enums_1.MathStyle.Display] = enums_1.MathStyle.Text,
        _a[enums_1.MathStyle.Text] = enums_1.MathStyle.Script,
        _a[enums_1.MathStyle.Script] = enums_1.MathStyle.ScriptScript,
        _a[enums_1.MathStyle.ScriptScript] = enums_1.MathStyle.ScriptScript,
        _a)[mathStyle];
    var childContext = __assign(__assign({}, context), { mathStyle: childMathStyle });
    return childContext;
};
var typesetTable = function (typesetChild, node, context, zipper) {
    var _a;
    var columns = [];
    var rows = [];
    var childContext = childContextForTable(context);
    var reboxColumn = function (col, leftKernSize, rightKernSize) {
        for (var row = 0; row < node.rowCount; row++) {
            var cell = rows[row].children[col];
            cell = Layout.rebox(cell, Layout.makeKern(leftKernSize), Layout.makeKern(rightKernSize));
            rows[row].children[col] = cell;
            columns[col].children[row] = cell;
            columns[col].width = Math.max(columns[col].width, cell.width);
        }
    };
    var gutterWidth = node.subtype === "algebra" ? 0 : DEFAULT_GUTTER_WIDTH;
    var children = node.type === "table"
        ? node.children
        : __spreadArray(__spreadArray(__spreadArray([], __read(node.left), false), [
            // @ts-expect-error: zipper is always defined when
            // node is a ZTable
            Editor.zrowToRow(zipper.row)
        ], false), __read(node.right), false);
    var MIN_WIDTH = 0; // 32; // only use this for debugging purposes.
    // Group cells into rows and columns and determine the width of each
    // column and the depth/height of each row.
    for (var j = 0; j < node.rowCount; j++) {
        for (var i = 0; i < node.colCount; i++) {
            if (!columns[i]) {
                columns[i] = {
                    children: [],
                    width: MIN_WIDTH,
                };
            }
            if (!rows[j]) {
                rows[j] = {
                    children: [],
                    height: 0,
                    depth: 0,
                };
            }
            // We want to add padding around the first operator if it's the only
            // character in the cell.
            var child = children[j * node.colCount + i];
            var padFirstOperator = ((_a = child === null || child === void 0 ? void 0 : child.children) === null || _a === void 0 ? void 0 : _a.length) === 1 &&
                child.children[0].type === "char" &&
                ["+", "\u2212"].includes(child.children[0].value);
            var cell = typesetChild(j * node.colCount + i, childContext, padFirstOperator);
            if (cell) {
                columns[i].width = Math.max(cell.width, columns[i].width);
                rows[j].height = Math.max(cell.height, rows[j].height);
                rows[j].depth = Math.max(cell.depth, rows[j].depth);
            }
            else {
                // Use an empty Layout.Box for children that were null
                cell = {
                    type: "HBox",
                    shift: 0,
                    content: {
                        type: "static",
                        nodes: [],
                    },
                    // These values don't matter since the box is empty
                    fontSize: 0,
                    style: {},
                    // These will get filled in later
                    width: 0,
                    height: 0,
                    depth: 0,
                };
            }
            columns[i].children.push(cell);
            rows[j].children.push(cell);
        }
    }
    var cursorIndex = node.type === "ztable" ? node.left.length : -1;
    if (node.subtype === "algebra" && cursorIndex !== -1) {
        var col = cursorIndex % node.colCount;
        var zrow = zipper === null || zipper === void 0 ? void 0 : zipper.row;
        var cellColumns = [];
        for (var i = 0; i < node.colCount; i++) {
            var col_1 = [];
            for (var j = 0; j < node.rowCount; j++) {
                var index = j * node.colCount + i;
                // TODO: check that children[index] isn't null
                col_1.push(children[index]);
            }
            cellColumns.push(col_1); // this is unsafe
        }
        if (zrow) {
            if (col === 0 && !cellColumns[col + 1].some(isOperator)) {
                // Add right padding on every cell in the first column
                reboxColumn(col, 0, 16);
            }
            else if (col === node.colCount - 1 &&
                !cellColumns[col - 1].some(isOperator)) {
                // Add left padding on every cell in the last column
                reboxColumn(col, 16, 0);
            }
            else if (col > 0 &&
                col < cellColumns.length - 1 &&
                Editor.isColumnEmpty(cellColumns[col])) {
                // If the cursor is in an empty column, only add padding to one
                // side of the column if there's an operator in one of the
                // columns. The padding goes on the opposite side of the column
                // with the operator since operators have their own built-in
                // padding.
                if (cellColumns[col - 1].some(isOperator) &&
                    !cellColumns[col + 1].some(isOperator)) {
                    reboxColumn(col, 0, 16);
                }
                else if (cellColumns[col + 1].some(isOperator) &&
                    !cellColumns[col - 1].some(isOperator)) {
                    reboxColumn(col, 16, 0);
                }
            }
        }
        // If there are any columns with no operators on either side, add both
        // left and right padding regardless of whether the cursor is in the
        // column or not.
        for (var col_2 = 0; col_2 < node.colCount; col_2++) {
            if (col_2 > 0 &&
                col_2 < cellColumns.length - 1 &&
                Editor.isColumnEmpty(cellColumns[col_2]) &&
                !cellColumns[col_2 - 1].some(isOperator) &&
                !cellColumns[col_2 + 1].some(isOperator)) {
                reboxColumn(col_2, 16, 16);
            }
        }
    }
    // Adjust the width of cells in the same column to be the same
    for (var i = 0; i < columns.length; i++) {
        var col = columns[i];
        for (var j = 0; j < col.children.length; j++) {
            // center the cell content
            var originalWidth = Layout.getWidth(col.children[j]);
            var baseKernSize = (col.width - originalWidth) / 2;
            var rightKernSize = i < columns.length - 1
                ? baseKernSize + gutterWidth / 2
                : baseKernSize;
            var leftKernSize = i > 0 ? baseKernSize + gutterWidth / 2 : baseKernSize;
            var child = children[j * node.colCount + i];
            // Right align any child that has content when showing work
            if (node.subtype === "algebra" &&
                child &&
                child.children.length > 0) {
                leftKernSize = baseKernSize * 2;
                rightKernSize = 0;
            }
            var cell = Layout.rebox(col.children[j], Layout.makeKern(leftKernSize, "start"), Layout.makeKern(rightKernSize, "end"));
            col.children[j] = cell;
            // rows[] has its own references so we need to update it as well
            rows[j].children[i] = cell;
        }
    }
    // Adjust the height/depth of cells in the same row to be the same
    for (var i = 0; i < rows.length; i++) {
        var row = rows[i];
        for (var j = 0; j < row.children.length; j++) {
            row.children[j].height = row.height;
            row.children[j].depth = row.depth;
        }
    }
    var width = columns.reduce(function (sum, col) { return sum + col.width; }, 0) +
        gutterWidth * (columns.length - 1);
    var constants = context.fontData.font.math.constants;
    var fontSize = (0, utils_1.fontSizeForContext)(context);
    var rowBoxes = rows.map(function (row, index) {
        var _a;
        var result = Layout.makeStaticHBox(row.children, context);
        var style = (_a = node.rowStyles) === null || _a === void 0 ? void 0 : _a[index];
        if ((style === null || style === void 0 ? void 0 : style.border) === "top") {
            var thickness = (fontSize * constants.fractionRuleThickness.value) / 1000;
            var stroke = Layout.makeStaticHBox([Layout.makeHRule(thickness, width)], context);
            var mathStyle = context.mathStyle;
            var useDisplayStyle = mathStyle === enums_1.MathStyle.Display;
            var minDenGap = useDisplayStyle
                ? (fontSize * constants.fractionDenomDisplayStyleGapMin.value) /
                    1000
                : (fontSize * constants.fractionDenominatorGapMin.value) / 1000;
            var minNumGap = useDisplayStyle
                ? (fontSize * constants.fractionNumDisplayStyleGapMin.value) /
                    1000
                : (fontSize * constants.fractionNumeratorGapMin.value) / 1000;
            return Layout.makeVBox(width, stroke, [Layout.makeKern(minNumGap)], [Layout.makeKern(minDenGap), result], context);
        }
        return result;
    });
    var inner = Layout.makeVBox(width, rowBoxes[0], [], rowBoxes.slice(1), context);
    var shift = (fontSize * constants.axisHeight.value) / 1000;
    // Equalize the depth and height and then shift up so the center of the
    // table aligns with the central axis.
    var vsize = Layout.vsize(inner);
    inner.height = vsize / 2 + shift;
    inner.depth = vsize / 2 - shift;
    var thresholdOptions = {
        value: "sum",
        strict: true,
    };
    if (!node.delimiters) {
        inner.id = node.id;
        inner.style = node.style;
        return inner;
    }
    var open = (0, utils_1.makeDelimiter)(node.delimiters.left.value, inner, thresholdOptions, context);
    var close = (0, utils_1.makeDelimiter)(node.delimiters.right.value, inner, thresholdOptions, context);
    var table = Layout.makeStaticHBox([open, inner, close], context);
    table.id = node.id;
    table.style = node.style;
    return table;
};
exports.typesetTable = typesetTable;
