"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeDelimiter = exports.fontSizeForContext = exports.multiplierForContext = void 0;
var enums_1 = require("./enums");
var Layout = __importStar(require("./layout"));
// TODO: return the font size instead of the multiplier
var multiplierForContext = function (context) {
    var constants = context.fontData.font.math.constants;
    var mathStyle = context.mathStyle;
    switch (mathStyle) {
        case enums_1.MathStyle.Display:
        case enums_1.MathStyle.Text:
            return 1.0;
        case enums_1.MathStyle.Script:
            return constants.scriptPercentScaleDown / 100;
        case enums_1.MathStyle.ScriptScript:
            return constants.scriptScriptPercentScaleDown / 100;
    }
};
exports.multiplierForContext = multiplierForContext;
var fontSizeForContext = function (context) {
    var multiplier = (0, exports.multiplierForContext)(context);
    var baseFontSize = context.baseFontSize;
    var fontSize = multiplier * baseFontSize;
    return fontSize;
};
exports.fontSizeForContext = fontSizeForContext;
// TODO: special case how we compute the delimiters for rows including "y" or
// other deep descenders so that we get the same font size as the rest of the
// glyphs on that row.
var getDelimiter = function (char, box, thresholdOptions, context) {
    var font = context.fontData.font;
    var glyphID = font.getGlyphID(char);
    var construction = font.math.variants.getVertGlyphConstruction(glyphID);
    if (!construction) {
        return glyphID;
    }
    var fontSize = (0, exports.fontSizeForContext)(context);
    for (var i = 0; i < construction.mathGlyphVariantRecords.length; i++) {
        var record = construction.mathGlyphVariantRecords[i];
        var glyphMetrics = font.getGlyphMetrics(record.variantGlyph);
        var height = (glyphMetrics.bearingY * fontSize) / font.head.unitsPerEm;
        var depth = ((glyphMetrics.height - glyphMetrics.bearingY) * fontSize) /
            font.head.unitsPerEm;
        var compare = thresholdOptions.strict
            ? function (a, b) { return a > b; }
            : function (a, b) { return a >= b; };
        // TODO: add an option to configure whether these inequalities are
        // strict or not.
        switch (thresholdOptions.value) {
            case "both": {
                if (compare(height, box.height) && compare(depth, box.depth)) {
                    // HACK: this is to ensure that we're using the same size
                    // glyph as the row when it contains deep descenders like "y"
                    if (i === 1 && char !== "\u221a") {
                        return glyphID;
                    }
                    return record.variantGlyph;
                }
                break;
            }
            case "sum": {
                if (compare(height + depth, box.height + box.depth)) {
                    // HACK: this is to ensure that we're using the same size
                    // glyph as the row when it contains deep descenders like "y"
                    if (i === 1 && char !== "\u221a") {
                        return glyphID;
                    }
                    return record.variantGlyph;
                }
                break;
            }
        }
    }
    if (construction.mathGlyphVariantRecords.length > 0) {
        // TODO: return a glyph assembly layout instead of the tallest delim
        return construction.mathGlyphVariantRecords[construction.mathGlyphVariantRecords.length - 1].variantGlyph;
    }
    return glyphID;
};
/**
 * Returns the smallest delimiter whose depth and height exceed the that of the
 * box passed in.  `thresholdOptions` controls how this comparison is made.
 *
 * @param {string} char
 * @param {Layout.Box} box layout encompassed by the delimiter
 * @param {ThresholdOptions} options
 * @param {Context} context
 * @returns {Layout.Glyph}
 */
var makeDelimiter = function (char, box, thresholdOptions, context) {
    var glyphID = getDelimiter(char, box, thresholdOptions, context);
    return Layout.makeGlyph(char, glyphID, context);
};
exports.makeDelimiter = makeDelimiter;
